@@ -1,13 +1,26 @@
-#include <Rcpp.h>
 #include <fstream>
 #include <iostream>
 #include <sstream>
+#include <vector>
+#include <string>
+#include <map>
+#include <set>
 
-using namespace Rcpp;
 #include "readstat.h"
 #include "haven_types.h"
 #include "tagged_na.h"
 
+#include "cpp11/strings.hpp"
+#include "cpp11/doubles.hpp"
+#include "cpp11/integers.hpp"
+#include "cpp11/r_string.hpp"
+#include "cpp11/list.hpp"
+#include "cpp11/raws.hpp"
+#include "cpp11/sexp.hpp"
+
+#include "cpp11/protect.hpp"
+#include "cpp11/function.hpp"
+
 double haven_double_value_udm(readstat_value_t value, readstat_variable_t* var, bool user_na) {
   if (readstat_value_is_tagged_missing(value)) {
     return make_tagged_na(tolower(readstat_value_tag(value)));
@@ -42,7 +55,7 @@ public:
 
   void add(const char* value, std::string label) {
     if (values_i_.size() > 0 || values_d_.size() > 0)
-      stop("Can't add string to integer/double labelset");
+      cpp11::stop("Can't add string to integer/double labelset");
 
     values_s_.push_back(value);
     labels_.push_back(label);
@@ -50,7 +63,7 @@ public:
 
   void add(int value, std::string label) {
     if (values_d_.size() > 0 || values_s_.size() > 0)
-      stop("Can't add integer to string/double labelset");
+      cpp11::stop("Can't add integer to string/double labelset");
 
     values_i_.push_back(value);
     labels_.push_back(label);
@@ -58,7 +71,7 @@ public:
 
   void add(double value, std::string label) {
     if (values_i_.size() > 0 || values_s_.size() > 0)
-      stop("Can't add double to integer/string labelset");
+      cpp11::stop("Can't add double to integer/string labelset");
 
     values_d_.push_back(value);
     labels_.push_back(label);
@@ -68,40 +81,40 @@ public:
     return labels_.size();
   }
 
-  RObject labels() const {
-    RObject out;
+  cpp11::sexp labels() const {
+    cpp11::sexp out;
 
     if (values_i_.size() > 0) {
       int n = values_i_.size();
-      IntegerVector values(n);
-      CharacterVector labels(n);
+      cpp11::writable::integers values(n);
+      cpp11::writable::strings labels(n);
 
       for (int i = 0; i < n; ++i) {
         values[i] = values_i_[i];
-        labels[i] = Rf_mkCharCE(labels_[i].c_str(), CE_UTF8);
+        labels[i] = labels_[i].c_str();
       }
 
       values.attr("names") = labels;
       out = values;
     } else if (values_d_.size() > 0) {
       int n = values_d_.size();
-      NumericVector values(n);
-      CharacterVector labels(n);
+      cpp11::writable::doubles values(n);
+      cpp11::writable::strings labels(n);
 
       for (int i = 0; i < n; ++i) {
         values[i] = values_d_[i];
-        labels[i] = Rf_mkCharCE(labels_[i].c_str(), CE_UTF8);
+        labels[i] = labels_[i].c_str();
       }
 
       values.attr("names") = labels;
       out = values;
     } else {
       int n = values_s_.size();
-      CharacterVector values(n), labels(n);
+      cpp11::writable::strings values(n), labels(n);
 
       for (int i = 0; i < n; ++i) {
-        values[i] = Rf_mkCharCE(values_s_[i].c_str(), CE_UTF8);
-        labels[i] = Rf_mkCharCE(labels_[i].c_str(), CE_UTF8);
+        values[i] = values_s_[i].c_str();
+        labels[i] = labels_[i].c_str();
       }
 
       values.attr("names") = labels;
@@ -115,11 +128,12 @@ public:
 // DfReader ------------------------------------------------------------------
 
 class DfReader {
-  FileType type_;
+  FileVendor vendor_;
+
   int nrows_, nrowsAlloc_;
   int ncols_;
-  List output_;
-  CharacterVector names_;
+  cpp11::writable::list output_;
+  cpp11::writable::strings names_;
   bool user_na_;
 
   std::vector<std::string> val_labels_;
@@ -127,15 +141,19 @@ class DfReader {
   std::vector<VarType> var_types_;
   std::vector<std::string> notes_;
 
-  // If empty, assume all will be read in
-  std::set<std::string> colsOnly_;
+  std::set<std::string> colsSkip_;
 
 public:
-  DfReader(FileType type, bool user_na = false): type_(type), nrows_(0), ncols_(0), user_na_(user_na) {
-  }
+  DfReader(FileExt ext, bool user_na = false) :
+    vendor_(extVendor(ext)),
+    nrows_(0),
+    ncols_(0),
+    output_(static_cast<R_xlen_t>(0)),
+    user_na_(user_na) { }
 
-  void restrictCols(const std::set<std::string>& cols) {
-    colsOnly_ = cols;
+  void skipCols(const std::vector<std::string>& cols) {
+    std::set<std::string> cols_set(cols.begin(), cols.end());
+    colsSkip_ = cols_set;
   }
 
   void setInfo(int obs_count, int var_count) {
@@ -147,17 +165,21 @@ public:
       nrowsAlloc_ = nrows_ = obs_count;
     }
 
-    ncols_ = colsOnly_.empty() ? var_count : colsOnly_.size();
+    if (var_count < 1) {
+      return; // sas7bcat has var_count = 0
+    }
 
-    output_ = List(ncols_);
-    names_ = CharacterVector(ncols_);
+    ncols_ = var_count - colsSkip_.size();
+
+    output_.resize(ncols_);
+    names_.resize(ncols_);
     val_labels_.resize(ncols_);
     var_types_.resize(ncols_);
   }
 
   void setMetadata(const char *file_label) {
     if (file_label != NULL && strcmp(file_label, "") != 0) {
-      output_.attr("label") = CharacterVector::create(Rf_mkCharCE(file_label, CE_UTF8));
+      output_.attr("label") = file_label;
     }
   }
 
@@ -171,33 +193,33 @@ public:
 
     const char* name = readstat_variable_get_name(variable);
 
-    if (!colsOnly_.empty() && colsOnly_.count(name) == 0) {
+    if (colsSkip_.count(name) > 0) {
       return READSTAT_HANDLER_SKIP_VARIABLE;
     }
 
     int var_index = readstat_variable_get_index_after_skipping(variable);
 
-    names_[var_index] = Rf_mkCharCE(name, CE_UTF8);
+    names_[var_index] = name;
 
     switch(readstat_variable_get_type(variable)) {
     case READSTAT_TYPE_STRING_REF:
     case READSTAT_TYPE_STRING:
-      output_[var_index] = CharacterVector(nrowsAlloc_);
+      output_[var_index] = cpp11::writable::strings(nrowsAlloc_);
       break;
     case READSTAT_TYPE_INT8:
     case READSTAT_TYPE_INT16:
     case READSTAT_TYPE_INT32:
     case READSTAT_TYPE_FLOAT:
     case READSTAT_TYPE_DOUBLE:
-      output_[var_index] = NumericVector(nrowsAlloc_);
+      output_[var_index] = cpp11::writable::doubles(nrowsAlloc_);
       break;
     }
 
-    RObject col = output_[var_index];
+    cpp11::sexp col(output_[var_index]);
 
     const char* var_label = readstat_variable_get_label(variable);
     if (var_label != NULL && strcmp(var_label, "") != 0) {
-      col.attr("label") = CharacterVector::create(Rf_mkCharCE(var_label, CE_UTF8));
+      col.attr("label") = var_label;
     }
 
     if (val_labels != NULL)
@@ -205,7 +227,7 @@ public:
 
     const char* var_format = readstat_variable_get_format(variable);
 
-    VarType var_type = numType(type_, var_format);
+    VarType var_type = numType(vendor_, var_format);
     // Rcout << name << ": " << var_format << " [" << var_type << "]\n";
     var_types_[var_index] = var_type;
     switch(var_type) {
@@ -213,11 +235,11 @@ public:
       col.attr("class") = "Date";
       break;
     case HAVEN_TIME:
-      col.attr("class") = CharacterVector::create("hms", "difftime");
+      col.attr("class") = {"hms", "difftime"};
       col.attr("units") = "secs";
       break;
     case HAVEN_DATETIME:
-      col.attr("class") = CharacterVector::create("POSIXct", "POSIXt");
+      col.attr("class") = {"POSIXct", "POSIXt"};
       col.attr("tzone") = "UTC";
       break;
     default:
@@ -231,15 +253,35 @@ public:
       case READSTAT_TYPE_STRING_REF:
       case READSTAT_TYPE_STRING:
       {
-        CharacterVector na_values(n_ranges);
+        cpp11::writable::strings na_values(R_xlen_t(0));
+        cpp11::writable::strings na_range(2);
+        bool has_range = false;
 
         for (int i = 0; i < n_ranges; ++i) {
-          readstat_value_t value = readstat_variable_get_missing_range_lo(variable, i);
-          const char* str_value = readstat_string_value(value);
-          na_values[0] = str_value == NULL ? NA_STRING : Rf_mkCharCE(str_value, CE_UTF8);
+          readstat_value_t
+            lo_value = readstat_variable_get_missing_range_lo(variable, i),
+            hi_value = readstat_variable_get_missing_range_hi(variable, i);
+
+          const char* lo = readstat_string_value(lo_value);
+          const char* hi = readstat_string_value(hi_value);
+
+          if (lo == hi) {
+            // single value
+            na_values.push_back(lo == NULL ? cpp11::r_string(NA_STRING) : cpp11::r_string(lo));
+          } else {
+            has_range = true;
+            // Can only ever be one range
+            na_range[0] = lo;
+            na_range[1] = hi;
+          }
         }
 
-        col.attr("na_values") = na_values;
+        if (na_values.size() > 0)
+          col.attr("na_values") = na_values;
+        if (has_range)
+          col.attr("na_range") = na_range;
+
+        col.attr("class") = {"haven_labelled_spss", "haven_labelled", "vctrs_vctr", "character"};
         break;
       }
       case READSTAT_TYPE_INT8:
@@ -249,7 +291,7 @@ public:
       case READSTAT_TYPE_DOUBLE:
       {
         std::vector<double> na_values;
-        NumericVector na_range(2);
+        cpp11::writable::doubles na_range(2);
         bool has_range = false;
 
         for (int i = 0; i < n_ranges; ++i) {
@@ -273,24 +315,24 @@ public:
         }
 
         if (na_values.size() > 0)
-          col.attr("na_values") = Rcpp::wrap(na_values);
+          col.attr("na_values") = na_values;
         if (has_range)
           col.attr("na_range") = na_range;
 
+        col.attr("class") = {"haven_labelled_spss", "haven_labelled", "vctrs_vctr", "double"};
       }
       }
-      col.attr("class") = CharacterVector::create("labelled_spss", "labelled");
     }
 
 
     // Store original format as attribute
     if (var_format != NULL && strcmp(var_format, "") != 0) {
-      col.attr(formatAttribute(type_)) = Rf_ScalarString(Rf_mkCharCE(var_format, CE_UTF8));
+      col.attr(formatAttribute(vendor_)) = var_format;
     }
 
     // Store original display width as attribute if it differs from the default
     int display_width = readstat_variable_get_display_width(variable);
-    if (type_ == HAVEN_SPSS && display_width != 8) {
+    if (vendor_ == HAVEN_SPSS && display_width != 8) {
       col.attr("display_width") = Rf_ScalarInteger(display_width);
     }
 
@@ -311,10 +353,17 @@ public:
     case READSTAT_TYPE_STRING_REF:
     case READSTAT_TYPE_STRING:
     {
-      CharacterVector col = output_[var_index];
+      cpp11::writable::strings col(output_[var_index]);
 
-      const char* str_value = readstat_string_value(value);
-      col[obs_index] = str_value == NULL ? NA_STRING : Rf_mkCharCE(str_value, CE_UTF8);
+      if (readstat_value_is_tagged_missing(value)) {
+        col[obs_index] = NA_STRING;
+      } else if (!user_na_ && readstat_value_is_defined_missing(value, variable)) {
+        col[obs_index] = NA_STRING;
+      } else if (readstat_value_is_system_missing(value)) {
+        col[obs_index] = NA_STRING;
+      } else {
+        col[obs_index] = cpp11::r_string(readstat_string_value(value));
+      }
       break;
     }
     case READSTAT_TYPE_INT8:
@@ -323,9 +372,9 @@ public:
     case READSTAT_TYPE_FLOAT:
     case READSTAT_TYPE_DOUBLE:
     {
-      NumericVector col = output_[var_index];
+      cpp11::writable::doubles col(output_[var_index]);
       double val = haven_double_value_udm(value, variable, user_na_);
-      col[obs_index] = adjustDatetimeToR(type_, var_type, val);
+      col[obs_index] = adjustDatetimeToR(vendor_, var_type, val);
       break;
     }
     }
@@ -365,22 +414,28 @@ public:
     nrowsAlloc_ = n;
 
     for (int i = 0; i < ncols_; ++i) {
-      Shield<SEXP> copy(Rf_lengthgets(output_[i], n));
+      cpp11::sexp copy(Rf_lengthgets(output_[i], n));
       Rf_copyMostAttrib(output_[i], copy);
       output_[i] = copy;
     }
   }
 
-  List output() {
+  void limitRows(long n) {
+    if (nrows_ > n) {
+      nrows_ = n;
+    }
+  }
+
+  cpp11::list output(const std::string& name_repair) {
     if (nrows_ != nrowsAlloc_)
       resizeCols(nrows_);
 
     for (int i = 0; i < output_.size(); ++i) {
-      RObject col = output_[i];
+      cpp11::sexp col(output_[i]);
 
       if (hasLabel(i)) {
-        if (col.attr("class") == R_NilValue) {
-          col.attr("class") = "labelled";
+        if (Rf_getAttrib(col, R_ClassSymbol) == R_NilValue) {
+          col.attr("class") = {"haven_labelled", "vctrs_vctr", Rf_type2char(TYPEOF(col))};
         }
         col.attr("labels") = label_sets_[val_labels_[i]].labels();
       }
@@ -388,9 +443,10 @@ public:
 
     int nNotes = notes_.size();
     if (nNotes > 0) {
-      CharacterVector notes(nNotes);
+      cpp11::writable::strings notes(nNotes);
+
       for (int i = 0; i < nNotes; ++i) {
-        notes[i] = Rf_mkCharCE(notes_[i].c_str(), CE_UTF8);
+        notes[i] = notes_[i].c_str();
       }
 
       output_.attr("notes") = notes_;
@@ -398,10 +454,10 @@ public:
 
     output_.attr("names") = names_;
 
-    static Function as_tibble("as_tibble", Environment::namespace_env("tibble"));
-    return as_tibble(output_);
+    static cpp11::function as_tibble = cpp11::package("tibble")["as_tibble"];
+    using namespace cpp11::literals;
+    return SEXP(as_tibble(output_, ".rows"_nm = nrows_, ".name_repair"_nm = name_repair));
   }
-
 };
 
 int dfreader_metadata(readstat_metadata_t *metadata, void *ctx) {
@@ -426,7 +482,7 @@ int dfreader_value(int obs_index, readstat_variable_t *variable,
                    readstat_value_t value, void *ctx) {
   // Check for user interrupts every 10,000 rows or cols
   if ((obs_index + 1) % 10000 == 0 || (variable->index + 1) % 10000 == 0)
-    checkUserInterrupt();
+    cpp11::check_user_interrupt();
 
   ((DfReader*) ctx)->setValue(obs_index, variable, value);
   return 0;
@@ -438,7 +494,7 @@ int dfreader_value_label(const char *val_labels, readstat_value_t value,
 }
 
 void print_error(const char* error_message, void* ctx) {
-  Rcout << error_message << "\n";
+  Rprintf("%s\n", error_message);
 }
 
 
@@ -452,6 +508,8 @@ public:
   virtual int close(void* io_ctx) = 0;
   virtual readstat_off_t seek(readstat_off_t offset, readstat_io_flags_t whence, void *io_ctx) = 0;
   virtual ssize_t read(void *buf, size_t nbyte, void *io_ctx) = 0;
+  virtual std::string source() const = 0; // human readable description of input source
+  std::string encoding;
 };
 
 template <typename Stream>
@@ -483,9 +541,14 @@ class DfReaderInputFile : public DfReaderInputStream<std::ifstream> {
   std::string filename_;
 
 public:
-  DfReaderInputFile(Rcpp::List spec) {
-    CharacterVector path(spec[0]);
+  DfReaderInputFile(cpp11::list spec, std::string encoding = "") {
+    cpp11::strings path(spec[0]);
     filename_ = std::string(Rf_translateChar(path[0]));
+    this->encoding = encoding;
+  }
+
+  std::string source() const {
+    return filename_;
   }
 
   int open(void* io_ctx) {
@@ -501,10 +564,15 @@ public:
 
 class DfReaderInputRaw : public DfReaderInputStream<std::istringstream> {
 public:
-  DfReaderInputRaw(Rcpp::List spec) {
-    Rcpp::RawVector raw_data(spec[0]);
+  DfReaderInputRaw(cpp11::list spec, std::string encoding = "") {
+    cpp11::raws raw_data(spec[0]);
     std::string string_data((char*) RAW(raw_data), Rf_length(raw_data));
     file_.str(string_data);
+    this->encoding = encoding;
+  }
+
+  std::string source() const {
+    return "file";
   }
 
   int open(void* io_ctx) {
@@ -535,7 +603,7 @@ readstat_error_t dfreader_update(long file_size, readstat_progress_handler progr
 
 // Parser wrappers -------------------------------------------------------------
 
-readstat_parser_t* haven_init_parser(std::string encoding = "") {
+readstat_parser_t* haven_init_parser() {
   readstat_parser_t* parser = readstat_parser_init();
   readstat_set_metadata_handler(parser, dfreader_metadata);
   readstat_set_note_handler(parser, dfreader_note);
@@ -543,194 +611,133 @@ readstat_parser_t* haven_init_parser(std::string encoding = "") {
   readstat_set_value_handler(parser, dfreader_value);
   readstat_set_value_label_handler(parser, dfreader_value_label);
   readstat_set_error_handler(parser, print_error);
-  if (encoding != "") {
-    readstat_set_file_character_encoding(parser, encoding.c_str());
-  }
   return parser;
 }
 
-template<typename InputClass>
-void haven_init_io(readstat_parser_t* parser, InputClass &builder_input) {
+void haven_init_io(readstat_parser_t* parser, DfReaderInput& builder_input) {
   readstat_set_open_handler(parser, dfreader_open);
   readstat_set_close_handler(parser, dfreader_close);
   readstat_set_seek_handler(parser, dfreader_seek);
   readstat_set_read_handler(parser, dfreader_read);
   readstat_set_update_handler(parser, dfreader_update);
   readstat_set_io_ctx(parser, (void*) &builder_input);
+  if (builder_input.encoding != "") {
+    readstat_set_file_character_encoding(parser, builder_input.encoding.c_str());
+  }
 }
 
-std::string haven_error_message(Rcpp::List spec) {
-  std::string source_class(as<CharacterVector>(spec.attr("class"))[0]);
-  if (source_class == "source_raw")
-    return "file";
-  else
-    return as<std::string>(spec[0]);
+void haven_set_row_limit(readstat_parser_t* parser, long n) {
+  // readstat uses 0 to specify "all rows" but what we want is "minimal rows"
+  readstat_set_row_limit(parser, n == 0 ? 1 : n);
 }
 
-template<typename InputClass>
-List df_parse_spss(Rcpp::List spec, std::string encoding = "", bool user_na = false, bool por = false) {
-  DfReader builder(HAVEN_SPSS, user_na);
-  InputClass builder_input(spec);
-
-  readstat_parser_t* parser = haven_init_parser(encoding);
+template<FileExt ext>
+void haven_parse(readstat_parser_t* parser, DfReaderInput& builder_input, DfReader* builder) {
   haven_init_io(parser, builder_input);
 
   readstat_error_t result;
-  if (por) {
-    result = readstat_parse_por(parser, "", &builder);
-  } else {
-    result = readstat_parse_sav(parser, "", &builder);
+  switch(ext) {
+  // the path for readstat_parse_* is baked into the io context (DfReaderInput)
+  case HAVEN_SAS7BDAT: result = readstat_parse_sas7bdat(parser, "", builder); break;
+  case HAVEN_SAS7BCAT: result = readstat_parse_sas7bcat(parser, "", builder); break;
+  case HAVEN_XPT:      result = readstat_parse_xport(parser, "", builder); break;
+  case HAVEN_DTA:      result = readstat_parse_dta(parser, "", builder); break;
+  case HAVEN_SAV:      result = readstat_parse_sav(parser, "", builder); break;
+  case HAVEN_POR:      result = readstat_parse_por(parser, "", builder); break;
+  default:             result = READSTAT_ERROR_PARSE; break;
   }
 
-  readstat_parser_free(parser);
-
-  if (result != 0) {
-    stop("Failed to parse %s: %s.", haven_error_message(spec),
-         readstat_error_message(result));
+  if (result != READSTAT_OK) {
+    std::string source = builder_input.source();
+    readstat_parser_free(parser);
+    std::string msg(readstat_error_message(result));
+    cpp11::stop("Failed to parse %s: %s.", source.c_str(), msg.c_str());
   }
-
-  return builder.output();
 }
 
-template<typename InputClass>
-List df_parse_dta(Rcpp::List spec, std::string encoding = "") {
-  DfReader builder(HAVEN_STATA);
-  InputClass builder_input(spec);
-
-  readstat_parser_t* parser = haven_init_parser(encoding);
-  haven_init_io(parser, builder_input);
-
-  readstat_error_t result = readstat_parse_dta(parser, "", &builder);
-  readstat_parser_free(parser);
-
-  if (result != 0) {
-    stop("Failed to parse %s: %s.",
-         haven_error_message(spec),
-         readstat_error_message(result));
-  }
-
-  return builder.output();
-}
-
-
-template<typename InputClass>
-List df_parse_xpt(Rcpp::List spec, std::string encoding = "") {
-  DfReader builder(HAVEN_XPT);
-  InputClass builder_input(spec);
-
-  readstat_parser_t* parser = haven_init_parser(encoding);
-  haven_init_io(parser, builder_input);
-
-  readstat_error_t result = readstat_parse_xport(parser, "", &builder);
-  readstat_parser_free(parser);
-
-  if (result != 0) {
-    stop("Failed to parse %s: %s.",
-         haven_error_message(spec),
-         readstat_error_message(result));
-  }
-
-  return builder.output();
-}
-
-
-template<typename InputClass>
-List df_parse_sas(Rcpp::List spec_b7dat, Rcpp::List spec_b7cat,
-                  std::string encoding, std::string catalog_encoding,
-                  std::vector<std::string> cols_only) {
-  DfReader builder(HAVEN_SAS);
-
-  if (!cols_only.empty()) {
-    std::set<std::string> cols_set(cols_only.begin(), cols_only.end());
-    builder.restrictCols(cols_set);
-  }
-
-  InputClass builder_input_dat(spec_b7dat);
+template<FileExt ext, typename InputClass>
+cpp11::list df_parse(cpp11::list spec, const std::vector<std::string>& cols_skip,
+              const long& n_max = -1, const long& rows_skip = 0,
+              const std::string& encoding = "",
+              const bool& user_na = false,
+              const std::string& name_repair = "check_unique",
+              cpp11::list catalog_spec = cpp11::writable::list(R_xlen_t(0)),
+              const std::string& catalog_encoding = ""
+              ) {
+  DfReader builder(ext, user_na);
+  builder.skipCols(cols_skip);
 
   readstat_parser_t* parser = haven_init_parser();
-  haven_init_io(parser, builder_input_dat);
+  haven_set_row_limit(parser, n_max);
+  readstat_set_row_offset(parser, rows_skip);
 
-  if (spec_b7cat.size() != 0) {
-    InputClass builder_input_cat(spec_b7cat);
-    readstat_set_io_ctx(parser, (void*) &builder_input_cat);
-    if (catalog_encoding != "") {
-      readstat_set_file_character_encoding(parser, catalog_encoding.c_str());
-    }
-
-    readstat_error_t result = readstat_parse_sas7bcat(parser, "", &builder);
-
-    if (result != 0) {
-      readstat_parser_free(parser);
-      stop("Failed to parse %s: %s.", haven_error_message(spec_b7cat),
-           readstat_error_message(result));
-    }
+  if (ext == HAVEN_SAS7BDAT && catalog_spec.size() != 0) {
+    InputClass cat_builder_input(catalog_spec, catalog_encoding);
+    haven_parse<HAVEN_SAS7BCAT>(parser, cat_builder_input, &builder);
   }
 
-  readstat_set_io_ctx(parser, (void*) &builder_input_dat);
-  if (encoding != "") {
-    readstat_set_file_character_encoding(parser, encoding.c_str());
-  }
-
-  readstat_error_t result = readstat_parse_sas7bdat(parser, "", &builder);
+  InputClass builder_input(spec, encoding);
+  haven_parse<ext>(parser, builder_input, &builder);
   readstat_parser_free(parser);
 
-  if (result != 0) {
-    stop("Failed to parse %s: %s.", haven_error_message(spec_b7dat),
-         readstat_error_message(result));
+  if (n_max >= 0) {
+    builder.limitRows(n_max); // must enforce n_max = 0
   }
 
-  return builder.output();
+  return builder.output(name_repair);
 }
 
 // # nocov start
 
-// [[Rcpp::export]]
-List df_parse_sas_file(Rcpp::List spec_b7dat, Rcpp::List spec_b7cat,
+[[cpp11::register]]
+cpp11::list df_parse_sas_file(cpp11::list spec_b7dat, cpp11::list spec_b7cat,
                        std::string encoding, std::string catalog_encoding,
-                       std::vector<std::string> cols_only) {
-  return df_parse_sas<DfReaderInputFile>(spec_b7dat, spec_b7cat, encoding, catalog_encoding, cols_only);
+                       std::vector<std::string> cols_skip, long n_max, long rows_skip,
+                       std::string name_repair) {
+  return df_parse<HAVEN_SAS7BDAT, DfReaderInputFile>(spec_b7dat, cols_skip, n_max, rows_skip, encoding, false, name_repair, spec_b7cat, catalog_encoding);
 }
-// [[Rcpp::export]]
-List df_parse_sas_raw(Rcpp::List spec_b7dat, Rcpp::List spec_b7cat,
+[[cpp11::register]]
+cpp11::list df_parse_sas_raw(cpp11::list spec_b7dat, cpp11::list spec_b7cat,
                       std::string encoding, std::string catalog_encoding,
-                      std::vector<std::string> cols_only) {
-  return df_parse_sas<DfReaderInputRaw>(spec_b7dat, spec_b7cat, encoding, catalog_encoding, cols_only);
+                      std::vector<std::string> cols_skip, long n_max, long rows_skip,
+                      std::string name_repair) {
+  return df_parse<HAVEN_SAS7BDAT, DfReaderInputRaw>(spec_b7dat, cols_skip, n_max, rows_skip, encoding, false, name_repair, spec_b7cat, catalog_encoding);
 }
 
-// [[Rcpp::export]]
-List df_parse_xpt_file(Rcpp::List spec) {
-  return df_parse_xpt<DfReaderInputFile>(spec);
+[[cpp11::register]]
+cpp11::list df_parse_xpt_file(cpp11::list spec, std::vector<std::string> cols_skip, long n_max, long rows_skip, std::string name_repair) {
+  return df_parse<HAVEN_XPT, DfReaderInputFile>(spec, cols_skip, n_max, rows_skip, "", false, name_repair);
 }
-// [[Rcpp::export]]
-List df_parse_xpt_raw(Rcpp::List spec) {
-  return df_parse_xpt<DfReaderInputRaw>(spec);
+[[cpp11::register]]
+cpp11::list df_parse_xpt_raw(cpp11::list spec, std::vector<std::string> cols_skip, long n_max, long rows_skip, std::string name_repair) {
+  return df_parse<HAVEN_XPT, DfReaderInputRaw>(spec, cols_skip, n_max, rows_skip, "", false, name_repair);
 }
 
-// [[Rcpp::export]]
-List df_parse_dta_file(Rcpp::List spec, std::string encoding) {
-  return df_parse_dta<DfReaderInputFile>(spec, encoding);
+[[cpp11::register]]
+cpp11::list df_parse_dta_file(cpp11::list spec, std::string encoding, std::vector<std::string> cols_skip, long n_max, long rows_skip, std::string name_repair) {
+  return df_parse<HAVEN_DTA, DfReaderInputFile>(spec, cols_skip, n_max, rows_skip, encoding, false, name_repair);
 }
-// [[Rcpp::export]]
-List df_parse_dta_raw(Rcpp::List spec, std::string encoding) {
-  return df_parse_dta<DfReaderInputRaw>(spec, encoding);
+[[cpp11::register]]
+cpp11::list df_parse_dta_raw(cpp11::list spec, std::string encoding, std::vector<std::string> cols_skip, long n_max, long rows_skip, std::string name_repair) {
+  return df_parse<HAVEN_DTA, DfReaderInputRaw>(spec, cols_skip, n_max, rows_skip, encoding, false, name_repair);
 }
 
-// [[Rcpp::export]]
-List df_parse_sav_file(Rcpp::List spec, std::string encoding, bool user_na) {
-  return df_parse_spss<DfReaderInputFile>(spec, encoding, user_na, false);
+[[cpp11::register]]
+cpp11::list df_parse_sav_file(cpp11::list spec, std::string encoding, bool user_na, std::vector<std::string> cols_skip, long n_max, long rows_skip, std::string name_repair) {
+  return df_parse<HAVEN_SAV, DfReaderInputFile>(spec, cols_skip, n_max, rows_skip, encoding, user_na, name_repair);
 }
-// [[Rcpp::export]]
-List df_parse_sav_raw(Rcpp::List spec, std::string encoding, bool user_na) {
-  return df_parse_spss<DfReaderInputRaw>(spec, encoding, user_na, false);
+[[cpp11::register]]
+cpp11::list df_parse_sav_raw(cpp11::list spec, std::string encoding, bool user_na, std::vector<std::string> cols_skip, long n_max, long rows_skip, std::string name_repair) {
+  return df_parse<HAVEN_SAV, DfReaderInputRaw>(spec, cols_skip, n_max, rows_skip, encoding, user_na, name_repair);
 }
 
-// [[Rcpp::export]]
-List df_parse_por_file(Rcpp::List spec, std::string encoding, bool user_na) {
-  return df_parse_spss<DfReaderInputFile>(spec, encoding, user_na, true);
+[[cpp11::register]]
+cpp11::list df_parse_por_file(cpp11::list spec, std::string encoding, bool user_na, std::vector<std::string> cols_skip, long n_max, long rows_skip, std::string name_repair) {
+  return df_parse<HAVEN_POR, DfReaderInputFile>(spec, cols_skip, n_max, rows_skip, encoding, user_na, name_repair);
 }
-// [[Rcpp::export]]
-List df_parse_por_raw(Rcpp::List spec, std::string encoding, bool user_na) {
-  return df_parse_spss<DfReaderInputRaw>(spec, encoding, user_na, true);
+[[cpp11::register]]
+cpp11::list df_parse_por_raw(cpp11::list spec, std::string encoding, bool user_na, std::vector<std::string> cols_skip, long n_max, long rows_skip, std::string name_repair) {
+  return df_parse<HAVEN_POR, DfReaderInputRaw>(spec, cols_skip, n_max, rows_skip, encoding, user_na, name_repair);
 }
 
 // # nocov end
