@@ -1,10 +1,10 @@
 
 #line 1 "src/txt/readstat_spss_commands_read.rl"
 #include <stdlib.h>
-#include <strings.h>
 #include <inttypes.h>
 
 #include "../readstat.h"
+#include "../readstat_strings.h"
 #include "readstat_schema.h"
 
 #include "readstat_copy.h"
@@ -1984,7 +1984,7 @@ _match:
 	break;
 	case 31:
 #line 259 "src/txt/readstat_spss_commands_read.rl"
-	{ label_type = LABEL_TYPE_DOUBLE; double_value = -integer; }
+	{ label_type = LABEL_TYPE_DOUBLE; double_value = -(double)integer; }
 	break;
 	case 32:
 #line 260 "src/txt/readstat_spss_commands_read.rl"
